/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***
	   
	 Vulkan Graphics API: in 20 Minutes (Coffee Break Series)
	 Beginners Taste

     Minimal working examples (MWE) to build upon and 
	 modify to gain an understanding of the API

	 Extensions to the Vulkan Graphics Coffee Break Series
	   
/*
/****************************************************************/

#pragma once

#pragma comment(linker, "/subsystem:windows")
#include <windows.h>


#include "vulkan.h"


#include "helpers.h"


#include <stdio.h>
#include <stdlib.h>

using namespace std;
#include <vector>
#include <array>


extern VkCoffeeDevice *				coffeeDevice;
extern VkFormat						colorformat;
extern VkFormat						depthFormat;
extern VkCoffeeSwapChain*			coffeeSwapChain;
extern VkCoffeeTexture*				coffeeTexture;


class CoffeeOffscreen
{
public:

	struct UniformData 
	{
		VkBuffer               buffer;
		VkDeviceMemory         memory;
		VkDescriptorBufferInfo descriptor;
		uint32_t               allocSize;
		void*                  mapped     = nullptr;
	} uniformData;

	struct FrameBufferAttachment 
	{
		VkImage        image;
		VkDeviceMemory mem;
		VkImageView    view;
	};
	
	struct 
	{
		VkCoffeeMesh*	cube;
		VkCoffeeMesh*	quad;
		VkCoffeeMesh*	plane;
	} meshes;

	struct UBO 
	{
		Matrix4 MVP;
	} ubo;


	VkDescriptorSet			descriptorSet; 
	VkDescriptorSetLayout	descriptorSetLayout;
	VkPipelineLayout		pipelineLayout;
	VkDescriptorPool		descriptorPool;
	VkCommandPool			cmdPool;

	int32_t					width;
	int32_t					height;
	VkFramebuffer			frameBuffer;		
	FrameBufferAttachment	color;
	FrameBufferAttachment	depth;
	VkRenderPass			renderPass;
	VkSampler				sampler;
	VkSampler				sampler2;
	VkDescriptorImageInfo	descriptorImageInfo;
	VkDescriptorImageInfo	descriptorImageInfo2;
	VkPipeline				pipelineShadedOffscreen;
	VkShaderModule			shaderModuleVS;
	VkShaderModule			shaderModulePS;
	VkCommandBuffer			commandBuffer		= VK_NULL_HANDLE;

	Matrix4					MVP;	// Model View Projection Transform


	CoffeeOffscreen()
	{
		meshes.cube  = new VkCoffeeMesh( coffeeDevice, VkCoffeeMesh::Cube  );
		meshes.quad  = new VkCoffeeMesh( coffeeDevice, VkCoffeeMesh::Quad  );
		meshes.plane = new VkCoffeeMesh( coffeeDevice, VkCoffeeMesh::Plane );

		CreateCommandPool();
		CreateCommandBuffers();

		SetupRenderPass();	

		PrepareUniformBuffersOffscreen();

		SetupDescriptorPool();
		SetupDescriptorSetLayout();

		
		PrepareOffscreenPipelines();	
	}

	void RenderOffscreen()
	{
		BuildOffscreenCommandBuffer();
		UpdateUniformBuffers();


		VkFence renderFence;
		VkFenceCreateInfo fenceCreateInfo = {};
		fenceCreateInfo.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;
		vkCreateFence( coffeeDevice->device, &fenceCreateInfo, NULL, &renderFence );
	
		VkSubmitInfo si = {};
		si.sType                = VK_STRUCTURE_TYPE_SUBMIT_INFO;
		si.waitSemaphoreCount   = 0;
		si.pWaitSemaphores      = VK_NULL_HANDLE;
		si.pWaitDstStageMask    = NULL;
		si.commandBufferCount   = 1;
		si.pCommandBuffers      = &commandBuffer;
		si.signalSemaphoreCount = 0;
		si.pSignalSemaphores    = VK_NULL_HANDLE;
	
		vkQueueSubmit( coffeeDevice->queue, 1, &si, renderFence );
	
		vkWaitForFences( coffeeDevice->device, 1, &renderFence, VK_TRUE, UINT64_MAX );
		vkDestroyFence( coffeeDevice->device, renderFence, NULL );
	}


	void UpdateUniformBuffers()
	{
		ubo.MVP = MVP;

		uint8_t *pData;
		VK_CHECK_RESULT(vkMapMemory(coffeeDevice->device, uniformData.memory, 0, sizeof(ubo), 0, (void **)&pData));
		memcpy(pData, &ubo, sizeof(ubo));
		vkUnmapMemory(coffeeDevice->device, uniformData.memory);
	}


	void SetupDescriptorPool()
	{
		VkDescriptorPoolSize descriptorPoolSize0		= {};
		descriptorPoolSize0.type						= VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
		descriptorPoolSize0.descriptorCount				= 6;

		VkDescriptorPoolSize descriptorPoolSize1		= {};
		descriptorPoolSize1.type						= VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
		descriptorPoolSize1.descriptorCount				= 2;

		std::vector<VkDescriptorPoolSize> poolSizes =
		{
			descriptorPoolSize0,
			descriptorPoolSize1
		};

		VkDescriptorPoolCreateInfo descriptorPoolInfo	= {};
		descriptorPoolInfo.sType						= VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
		descriptorPoolInfo.pNext						= NULL;
		descriptorPoolInfo.poolSizeCount				= poolSizes.size();
		descriptorPoolInfo.pPoolSizes					= poolSizes.data();
		descriptorPoolInfo.maxSets						= 5;


		VK_CHECK_RESULT(vkCreateDescriptorPool(coffeeDevice->device, &descriptorPoolInfo, nullptr, &descriptorPool));
	}

	void CreateCommandPool()
	{
		VkCommandPoolCreateInfo cmdPoolInfo				= {};
		cmdPoolInfo.sType								= VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
		cmdPoolInfo.queueFamilyIndex					= coffeeSwapChain->queueNodeIndex;
		cmdPoolInfo.flags								= VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT;
		VK_CHECK_RESULT(vkCreateCommandPool(coffeeDevice->device, &cmdPoolInfo, nullptr, &cmdPool));
	}

	void CreateCommandBuffers()
	{
		VkCommandBufferAllocateInfo cmdBufAllocateInfo   = {};
		cmdBufAllocateInfo.sType                         = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
		cmdBufAllocateInfo.commandPool                   = cmdPool;
		cmdBufAllocateInfo.level                         = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
		cmdBufAllocateInfo.commandBufferCount            = 1;

		VK_CHECK_RESULT(vkAllocateCommandBuffers(coffeeDevice->device, &cmdBufAllocateInfo, &commandBuffer));
	}

	void SetupRenderPass()
	{
		width  = 1024;
		height = 1024;

		// Color attachment
		VkImageCreateInfo image							= {};
		image.sType										= VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
		image.pNext										= NULL;
		image.imageType									= VK_IMAGE_TYPE_2D;
		image.format									= VK_FORMAT_R8G8B8A8_UNORM;
		image.extent.width								= width;
		image.extent.height								= height;
		image.extent.depth								= 1;
		image.mipLevels									= 1;
		image.arrayLayers								= 1;
		image.samples									= VK_SAMPLE_COUNT_1_BIT;
		image.tiling									= VK_IMAGE_TILING_OPTIMAL;
		//  Sample directly from the color attachment
		image.usage										= VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_SAMPLED_BIT;

		VkMemoryAllocateInfo memAlloc					= {};
		memAlloc.sType									= VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
		memAlloc.pNext									= NULL;
		memAlloc.allocationSize							= 0;
		memAlloc.memoryTypeIndex						= 0;
		VkMemoryRequirements memReqs;

		VK_CHECK_RESULT(vkCreateImage(coffeeDevice->device, &image, nullptr, &color.image));
		vkGetImageMemoryRequirements(coffeeDevice->device, color.image, &memReqs);
		memAlloc.allocationSize = memReqs.size;
		memAlloc.memoryTypeIndex = coffeeDevice->GetMemoryType(memReqs.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);
		VK_CHECK_RESULT(vkAllocateMemory(coffeeDevice->device, &memAlloc, nullptr, &color.mem));
		VK_CHECK_RESULT(vkBindImageMemory(coffeeDevice->device, color.image, color.mem, 0));

		VkImageViewCreateInfo colorImageView = {};
		colorImageView.sType                            = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
		colorImageView.pNext                            = NULL;
		colorImageView.viewType                         = VK_IMAGE_VIEW_TYPE_2D;
		colorImageView.format                           = VK_FORMAT_R8G8B8A8_UNORM;
		colorImageView.subresourceRange                 = {};
		colorImageView.subresourceRange.aspectMask      = VK_IMAGE_ASPECT_COLOR_BIT;
		colorImageView.subresourceRange.baseMipLevel    = 0;
		colorImageView.subresourceRange.levelCount      = 1;
		colorImageView.subresourceRange.baseArrayLayer  = 0;
		colorImageView.subresourceRange.layerCount      = 1;
		colorImageView.image                            = color.image;
		VK_CHECK_RESULT(vkCreateImageView(coffeeDevice->device, &colorImageView, nullptr, &color.view));

		// Create sampler to sample from the attachment in the fragment shader
		VkSamplerCreateInfo samplerInfo					 = {};
		samplerInfo.sType								 = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
		samplerInfo.pNext								 = NULL;
		samplerInfo.magFilter							 = VK_FILTER_LINEAR;
		samplerInfo.minFilter							 = VK_FILTER_LINEAR;
		samplerInfo.mipmapMode							 = VK_SAMPLER_MIPMAP_MODE_LINEAR;
		samplerInfo.addressModeU						 = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
		samplerInfo.addressModeV						 = samplerInfo.addressModeU;
		samplerInfo.addressModeW						 = samplerInfo.addressModeU;
		samplerInfo.mipLodBias							 = 0.0f;
		samplerInfo.maxAnisotropy						 = 0;
		samplerInfo.minLod								 = 0.0f;
		samplerInfo.maxLod								 = 1.0f;
		samplerInfo.borderColor           = VK_BORDER_COLOR_FLOAT_OPAQUE_WHITE;
		VK_CHECK_RESULT(vkCreateSampler(coffeeDevice->device, &samplerInfo, nullptr, &sampler));

		// Depth stencil attachment
		image.format = depthFormat; // ************ important!!! remember to set `VK_IMAGE_USAGE_SAMPLED_BIT'
		image.usage  = VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT | VK_IMAGE_USAGE_SAMPLED_BIT; //VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT;

		VK_CHECK_RESULT(vkCreateImage(coffeeDevice->device, &image, nullptr, &depth.image));
		vkGetImageMemoryRequirements(coffeeDevice->device, depth.image, &memReqs);
		memAlloc.allocationSize = memReqs.size;
		memAlloc.memoryTypeIndex = coffeeDevice->GetMemoryType(memReqs.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);
		VK_CHECK_RESULT(vkAllocateMemory(coffeeDevice->device, &memAlloc, nullptr, &depth.mem));
		VK_CHECK_RESULT(vkBindImageMemory(coffeeDevice->device, depth.image, depth.mem, 0));

		VkImageViewCreateInfo depthStencilView           = {};
		depthStencilView.sType                           = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
		depthStencilView.pNext                           = NULL;
		depthStencilView.viewType                        = VK_IMAGE_VIEW_TYPE_2D;
		depthStencilView.format                          = depthFormat;
		depthStencilView.flags                           = 0;
		depthStencilView.subresourceRange                = {};
		depthStencilView.subresourceRange.aspectMask     = VK_IMAGE_ASPECT_DEPTH_BIT | VK_IMAGE_ASPECT_STENCIL_BIT;
		depthStencilView.subresourceRange.baseMipLevel   = 0;
		depthStencilView.subresourceRange.levelCount     = 1;
		depthStencilView.subresourceRange.baseArrayLayer = 0;
		depthStencilView.subresourceRange.layerCount     = 1;
		depthStencilView.image                           = depth.image;
		VK_CHECK_RESULT(vkCreateImageView(coffeeDevice->device, &depthStencilView, nullptr, &depth.view));

		// Create a separate render pass for the offscreen rendering as it may differ from the one used for scene rendering

		std::array<VkAttachmentDescription, 2> attchmentDescriptions = {};
		// Color attachment
		attchmentDescriptions[0].format                  = VK_FORMAT_R8G8B8A8_UNORM;
		attchmentDescriptions[0].samples                 = VK_SAMPLE_COUNT_1_BIT;
		attchmentDescriptions[0].loadOp                  = VK_ATTACHMENT_LOAD_OP_CLEAR;
		attchmentDescriptions[0].storeOp                 = VK_ATTACHMENT_STORE_OP_STORE;
		attchmentDescriptions[0].stencilLoadOp           = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
		attchmentDescriptions[0].stencilStoreOp          = VK_ATTACHMENT_STORE_OP_DONT_CARE;
		attchmentDescriptions[0].initialLayout           = VK_IMAGE_LAYOUT_UNDEFINED;
		attchmentDescriptions[0].finalLayout             = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
		// Depth attachment						         
		attchmentDescriptions[1].format                  = depthFormat;
		attchmentDescriptions[1].samples                 = VK_SAMPLE_COUNT_1_BIT;
		attchmentDescriptions[1].loadOp                  = VK_ATTACHMENT_LOAD_OP_CLEAR;
		attchmentDescriptions[1].storeOp                 = VK_ATTACHMENT_STORE_OP_STORE;//VK_ATTACHMENT_STORE_OP_DONT_CARE;
		attchmentDescriptions[1].stencilLoadOp           = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
		attchmentDescriptions[1].stencilStoreOp          = VK_ATTACHMENT_STORE_OP_DONT_CARE;
		attchmentDescriptions[1].initialLayout           = VK_IMAGE_LAYOUT_UNDEFINED;
		attchmentDescriptions[1].finalLayout             = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;

		VkAttachmentReference colorReference = { 0, VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL };
		VkAttachmentReference depthReference = { 1, VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL };

		VkSubpassDescription subpassDescription          = {};
		subpassDescription.pipelineBindPoint             = VK_PIPELINE_BIND_POINT_GRAPHICS;
		subpassDescription.colorAttachmentCount          = 1;
		subpassDescription.pColorAttachments             = &colorReference;
		subpassDescription.pDepthStencilAttachment       = &depthReference;

		// Use subpass dependencies for layout transitions
		std::array<VkSubpassDependency, 2> dependencies;

		dependencies[0].srcSubpass                       = VK_SUBPASS_EXTERNAL;
		dependencies[0].dstSubpass                       = 0;
		dependencies[0].srcStageMask                     = VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT;
		dependencies[0].dstStageMask                     = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
		dependencies[0].srcAccessMask                    = VK_ACCESS_MEMORY_READ_BIT;
		dependencies[0].dstAccessMask                    = VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
		dependencies[0].dependencyFlags                  = VK_DEPENDENCY_BY_REGION_BIT;
										                 
		dependencies[1].srcSubpass                       = 0;
		dependencies[1].dstSubpass                       = VK_SUBPASS_EXTERNAL;
		dependencies[1].srcStageMask                     = VK_PIPELINE_STAGE_LATE_FRAGMENT_TESTS_BIT;//VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
		dependencies[1].dstStageMask                     = VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT;
		dependencies[1].srcAccessMask                    = VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
		dependencies[1].dstAccessMask                    = VK_ACCESS_MEMORY_READ_BIT;
		dependencies[1].dependencyFlags                  = VK_DEPENDENCY_BY_REGION_BIT;



		// Create the actual renderpass
		VkRenderPassCreateInfo renderPassInfo            = {};
		renderPassInfo.sType						     = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO;
		renderPassInfo.attachmentCount				     = static_cast<uint32_t>(attchmentDescriptions.size());
		renderPassInfo.pAttachments					     = attchmentDescriptions.data();
		renderPassInfo.subpassCount					     = 1;
		renderPassInfo.pSubpasses					     = &subpassDescription;
		renderPassInfo.dependencyCount				     = static_cast<uint32_t>(dependencies.size());
		renderPassInfo.pDependencies				     = dependencies.data();

		VK_CHECK_RESULT(vkCreateRenderPass(coffeeDevice->device, &renderPassInfo, nullptr, &renderPass));

		VkImageView attachments[2];
		attachments[0] = color.view;
		attachments[1] = depth.view;

		VkFramebufferCreateInfo fbufCreateInfo           = {};
		fbufCreateInfo.sType                             = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
		fbufCreateInfo.pNext                             = NULL;
		fbufCreateInfo.renderPass                        = renderPass;
		fbufCreateInfo.attachmentCount                   = 2;
		fbufCreateInfo.pAttachments                      = attachments;
		fbufCreateInfo.width                             = width;
		fbufCreateInfo.height                            = height;
		fbufCreateInfo.layers                            = 1;

		VK_CHECK_RESULT(vkCreateFramebuffer(coffeeDevice->device, &fbufCreateInfo, nullptr, &frameBuffer));

		// Fill a descriptor for later use in a descriptor set 
		descriptorImageInfo.imageLayout                  = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
		descriptorImageInfo.imageView                    = color.view;
		descriptorImageInfo.sampler                      = sampler;

		descriptorImageInfo2.imageLayout                  = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
		descriptorImageInfo2.imageView                    = depth.view;
		descriptorImageInfo2.sampler                      = sampler;
	}



	// Sets up the command buffer that renders the scene to the offscreen frame buffer
	void BuildOffscreenCommandBuffer()
	{
		DBG_ASSERT(commandBuffer);

		VkCommandBufferBeginInfo cmdBufInfo            = {};
		cmdBufInfo.sType                               = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
		cmdBufInfo.pNext                               = NULL;

		VkClearValue clearValues[2];
		clearValues[0].color                           = { { 0.0f, 0.0f, 0.0f, 0.0f } };
		clearValues[1].depthStencil                    = { 1.0f, 0 };

		VkRenderPassBeginInfo renderPassBeginInfo      = {};
		renderPassBeginInfo.sType                      = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO;
		renderPassBeginInfo.pNext                      = NULL;
		renderPassBeginInfo.renderPass                 = renderPass;
		renderPassBeginInfo.framebuffer                = frameBuffer;
		renderPassBeginInfo.renderArea.extent.width    = width;
		renderPassBeginInfo.renderArea.extent.height   = height;
		renderPassBeginInfo.clearValueCount            = 2;
		renderPassBeginInfo.pClearValues               = clearValues;

		VK_CHECK_RESULT(vkBeginCommandBuffer(commandBuffer, &cmdBufInfo));

		vkCmdBeginRenderPass(commandBuffer, &renderPassBeginInfo, VK_SUBPASS_CONTENTS_INLINE);

		VkViewport viewport                            = {};
		viewport.width                                 = (float)width;
		viewport.height                                = (float)height;
		viewport.minDepth                              = 0;
		viewport.maxDepth                              = 1;

		vkCmdSetViewport(commandBuffer, 0, 1, &viewport);

		VkRect2D scissor                               = {};
		scissor.extent.width                           = width;
		scissor.extent.height                          = height;
		scissor.offset.x                               = 0;
		scissor.offset.y                               = 0;
		vkCmdSetScissor(commandBuffer, 0, 1, &scissor);

		VkDeviceSize offsets[1] = { 0 };

		// Plane
		vkCmdBindDescriptorSets(commandBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineLayout, 0, 1, &descriptorSet, 0, NULL);
		vkCmdBindPipeline(commandBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineShadedOffscreen);

		vkCmdBindVertexBuffers(commandBuffer, VERTEX_BUFFER_BIND_ID, 1, &meshes.plane->vbuf, offsets);
		vkCmdBindIndexBuffer(commandBuffer, meshes.plane->ibuf, 0, VK_INDEX_TYPE_UINT32);
		vkCmdDrawIndexed(commandBuffer, meshes.plane->indexCount, 1, 0, 0, 0);

		// Model
		vkCmdBindDescriptorSets(commandBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineLayout, 0, 1, &descriptorSet, 0, NULL);
		vkCmdBindPipeline(commandBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineShadedOffscreen);

		vkCmdBindVertexBuffers(commandBuffer, VERTEX_BUFFER_BIND_ID, 1, &meshes.cube->vbuf, offsets);
		vkCmdBindIndexBuffer(commandBuffer, meshes.cube->ibuf, 0, VK_INDEX_TYPE_UINT32);
		vkCmdDrawIndexed(commandBuffer, meshes.cube->indexCount, 1, 0, 0, 0);

		vkCmdEndRenderPass(commandBuffer);

		VK_CHECK_RESULT(vkEndCommandBuffer(commandBuffer));
	}

	VkDescriptorSetLayoutCreateInfo DescriptorSetLayoutCreateInfo(
		const VkDescriptorSetLayoutBinding* pBindings,
		uint32_t bindingCount)
	{
		VkDescriptorSetLayoutCreateInfo descriptorSetLayoutCreateInfo = {};
		descriptorSetLayoutCreateInfo.sType              = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
		descriptorSetLayoutCreateInfo.pNext              = NULL;
		descriptorSetLayoutCreateInfo.pBindings          = pBindings;
		descriptorSetLayoutCreateInfo.bindingCount       = bindingCount;
		return descriptorSetLayoutCreateInfo;
	}


	VkPipelineLayoutCreateInfo PipelineLayoutCreateInfo(
		const VkDescriptorSetLayout* pSetLayouts,
		uint32_t setLayoutCount)
	{
		VkPipelineLayoutCreateInfo pipelineLayoutCreateInfo = {};
		pipelineLayoutCreateInfo.sType                  = VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO;
		pipelineLayoutCreateInfo.pNext                  = NULL;
		pipelineLayoutCreateInfo.setLayoutCount         = setLayoutCount;
		pipelineLayoutCreateInfo.pSetLayouts            = pSetLayouts;
		return pipelineLayoutCreateInfo;
	}


	void SetupDescriptorSetLayout()
	{
		std::vector<VkDescriptorSetLayoutBinding> setLayoutBindings;
		VkDescriptorSetLayoutCreateInfo descriptorLayoutInfo;
		VkPipelineLayoutCreateInfo pipelineLayoutInfo;

		VkDescriptorSetLayoutBinding setLayoutBinding0  = {};
		setLayoutBinding0.descriptorType				= VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
		setLayoutBinding0.stageFlags					= VK_SHADER_STAGE_VERTEX_BIT;
		setLayoutBinding0.binding						= 0;
		setLayoutBinding0.descriptorCount				= 1; 

		VkDescriptorSetLayoutBinding setLayoutBinding1  = {};
		setLayoutBinding1.descriptorType				= VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
		setLayoutBinding1.stageFlags					= VK_SHADER_STAGE_FRAGMENT_BIT;
		setLayoutBinding1.binding						= 1;
		setLayoutBinding1.descriptorCount				= 1; 

		// Binding 0 : Vertex shader uniform buffer
		setLayoutBindings.push_back(setLayoutBinding0);
		// Binding 1 : Fragment shader image sampler
		setLayoutBindings.push_back(setLayoutBinding1);

		// Layouts
		descriptorLayoutInfo = DescriptorSetLayoutCreateInfo(setLayoutBindings.data(), setLayoutBindings.size());
		VK_CHECK_RESULT(vkCreateDescriptorSetLayout(coffeeDevice->device, &descriptorLayoutInfo, nullptr, &descriptorSetLayout));

		pipelineLayoutInfo = PipelineLayoutCreateInfo(&descriptorSetLayout, 1);
		VK_CHECK_RESULT(vkCreatePipelineLayout(coffeeDevice->device, &pipelineLayoutInfo, nullptr, &pipelineLayout));


		// texture descriptor set
		VkDescriptorSetAllocateInfo allocInfo			= {};
		allocInfo.sType									= VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
		allocInfo.pNext									= NULL;
		allocInfo.descriptorPool						= descriptorPool;
		allocInfo.pSetLayouts							= &descriptorSetLayout;
		allocInfo.descriptorSetCount					= 1;

		VK_CHECK_RESULT(vkAllocateDescriptorSets(coffeeDevice->device, &allocInfo, &descriptorSet));


		VkWriteDescriptorSet wd[2];
		wd[0].sType										= VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
		wd[0].dstSet									= descriptorSet;
		wd[0].dstBinding								= 0;
		wd[0].dstArrayElement							= 0;
		wd[0].descriptorCount							= 1;
		wd[0].descriptorType							= VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
		wd[0].pImageInfo								= NULL;
		wd[0].pBufferInfo								= &uniformData.descriptor;
		wd[0].pTexelBufferView							= NULL;

		wd[1].sType										= VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
		wd[1].dstSet									= descriptorSet;
		wd[1].dstBinding								= 1;
		wd[1].dstArrayElement							= 0;
		wd[1].descriptorCount							= 1;
		wd[1].descriptorType							= VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
		wd[1].pImageInfo								= &coffeeTexture->descriptorImageInfo;
		//wd[1].pImageInfo								  = &descriptorImageInfo;
		wd[1].pBufferInfo								= NULL;
		wd[1].pTexelBufferView							= NULL;

		vkUpdateDescriptorSets( coffeeDevice->device, 2, wd, 0, NULL );
	}




	void PrepareOffscreenPipelines()
	{
		VkPipelineInputAssemblyStateCreateInfo inputAssemblyState = {};
		inputAssemblyState.sType						= VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO;
		inputAssemblyState.topology						= VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST;
		inputAssemblyState.flags						= 0;
		inputAssemblyState.primitiveRestartEnable		= VK_FALSE;

		VkPipelineRasterizationStateCreateInfo rasterizationState = {};
		rasterizationState.sType						= VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO;
		rasterizationState.polygonMode					= VK_POLYGON_MODE_FILL;
		rasterizationState.cullMode						= VK_CULL_MODE_BACK_BIT;
		rasterizationState.frontFace					= VK_FRONT_FACE_COUNTER_CLOCKWISE;//VK_FRONT_FACE_CLOCKWISE;
		rasterizationState.flags						= 0;
		rasterizationState.depthClampEnable				= VK_FALSE;
		rasterizationState.lineWidth					= 1.0f;

		VkPipelineColorBlendAttachmentState blendAttachmentState = {};
		blendAttachmentState.colorWriteMask				= 0xf;
		blendAttachmentState.blendEnable				= VK_FALSE;

		VkPipelineColorBlendStateCreateInfo colorBlendState = {};
		colorBlendState.sType							= VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO;
		colorBlendState.pNext							= NULL;
		colorBlendState.attachmentCount					= 1;
		colorBlendState.pAttachments					= &blendAttachmentState;

		VkPipelineDepthStencilStateCreateInfo depthStencilState = {};
		depthStencilState.sType							= VK_STRUCTURE_TYPE_PIPELINE_DEPTH_STENCIL_STATE_CREATE_INFO;
		depthStencilState.depthTestEnable				= VK_TRUE;
		depthStencilState.depthWriteEnable				= VK_TRUE;
		depthStencilState.depthCompareOp				= VK_COMPARE_OP_LESS_OR_EQUAL;
		depthStencilState.front							= depthStencilState.back;
		depthStencilState.back.compareOp				= VK_COMPARE_OP_ALWAYS;

		VkPipelineViewportStateCreateInfo viewportState = {};
		viewportState.sType								= VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO;
		viewportState.viewportCount						= 1;
		viewportState.scissorCount						= 1;
		viewportState.flags								= 0;

		VkPipelineMultisampleStateCreateInfo multisampleState = {};
		multisampleState.sType							= VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO;
		multisampleState.rasterizationSamples			= VK_SAMPLE_COUNT_1_BIT;

		std::vector<VkDynamicState> dynamicStateEnables = {
			VK_DYNAMIC_STATE_VIEWPORT,
			VK_DYNAMIC_STATE_SCISSOR
		};

		VkPipelineDynamicStateCreateInfo dynamicState	= {};
		dynamicState.sType								= VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO;
		dynamicState.pDynamicStates						= dynamicStateEnables.data(),
		dynamicState.dynamicStateCount					= dynamicStateEnables.size();

		// Rendering pipeline
	
		VkGraphicsPipelineCreateInfo pipelineCreateInfo = {};
		pipelineCreateInfo.sType						= VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO;
		pipelineCreateInfo.pNext						= NULL;
		pipelineCreateInfo.layout						= pipelineLayout;
		pipelineCreateInfo.renderPass					= renderPass;
		pipelineCreateInfo.flags						= 0;

		// Load shaders
		std::array<VkPipelineShaderStageCreateInfo, 2> shaderStages;

		pipelineCreateInfo.pVertexInputState			= &meshes.cube->verticeDescription.inputState;
		pipelineCreateInfo.pInputAssemblyState			= &inputAssemblyState;
		pipelineCreateInfo.pRasterizationState			= &rasterizationState;
		pipelineCreateInfo.pColorBlendState				= &colorBlendState;
		pipelineCreateInfo.pMultisampleState			= &multisampleState;
		pipelineCreateInfo.pViewportState				= &viewportState;
		pipelineCreateInfo.pDepthStencilState			= &depthStencilState;
		pipelineCreateInfo.pDynamicState				= &dynamicState;
		pipelineCreateInfo.stageCount					= shaderStages.size();
		pipelineCreateInfo.pStages						= shaderStages.data();
		pipelineCreateInfo.renderPass					= renderPass;

		shaderStages[0] = LoadShader("./shadow.vert.spv", VK_SHADER_STAGE_VERTEX_BIT,   &shaderModuleVS);
		shaderStages[1] = LoadShader("./shadow.frag.spv", VK_SHADER_STAGE_FRAGMENT_BIT, &shaderModulePS);

		VK_CHECK_RESULT(vkCreateGraphicsPipelines(coffeeDevice->device, VK_NULL_HANDLE, 1, &pipelineCreateInfo, nullptr, &pipelineShadedOffscreen));
	}

	void PrepareUniformBuffersOffscreen()
	{
		// Offscreen vertex shader uniform buffer block 
		CreateBuffer(
			VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT,
			sizeof(ubo),
			nullptr,
			&uniformData.buffer,
			&uniformData.memory,
			&uniformData.descriptor);
	}



};







